from abc import ABC, abstractmethod
from os import PathLike
from pathlib import Path
from typing import Union

from cx_studio.utils import detect_encoding


class SubtitleLoader(ABC):
    def __init__(self, filename: Union[str | PathLike], encoding=None):
        self._filename = Path(filename)
        self._encoding = str(encoding)

    @property
    def encoding(self):
        if self._encoding and self._encoding != 'auto':
            return self._encoding
        return detect_encoding(self._filename)

    @abstractmethod
    def open(self):
        return True

    @abstractmethod
    def close(self):
        return True

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    @abstractmethod
    def subtitles(self):
        pass



from dataclasses import dataclass
from functools import lru_cache

from ..core import DataPackage, Time


@dataclass(order=True, frozen=True)
class StaticSubtitle:
    start: Time
    end: Time
    content: str
    attachment: DataPackage = None

    @lru_cache
    def duration(self):
        return self.end - self.start



import re
from dataclasses import dataclass
from enum import StrEnum
from typing import Union

from .cx_time import Time
from .cx_timebase import Timebase


class TCMode(StrEnum):
    Code = "timecode"
    Stamp = "timestamp"


TC_PATTERN = re.compile(r"(\d\d):(\d\d):(\d\d)([:;,\.])(\d\d\d?)")


@dataclass
class TimeCodeInfo:
    hh: int = 0
    mm: int = 0
    ss: int = 0
    ff: int = 0
    mode: TCMode = TCMode.Code
    timebase: Timebase = None

    @staticmethod
    def is_tc(s: str) -> bool:
        s = str(s)
        return TC_PATTERN.fullmatch(s) is not None

    @classmethod
    def from_timecode(cls, tc: str):
        if not isinstance(tc, str):
            raise TypeError("TimeCodeInfo can only parse timecode from str.")
        match = TC_PATTERN.fullmatch(tc.strip())
        if match is None:
            return TimeCodeInfo()
        groups = match.groups()
        sep = groups[3]
        parts = [int(groups[x]) for x in [0, 1, 2, 4]]
        is_frame = len(groups[4]) == 2
        return TimeCodeInfo(
            hh=parts[0],
            mm=parts[1],
            ss=parts[2],
            ff=parts[3],
            timebase=Timebase(frame_rate=-1, drop_frame=sep == ";"),
            mode=TCMode.Code if is_frame else TCMode.Stamp,
        )

    @classmethod
    def from_time(
            cls, t: Time, tc_mode: TCMode = TCMode.Code,
            timebase: Timebase = None
    ):
        if not isinstance(t, Time):
            raise TypeError("TimeCodeInfo can only handle time by Time type.")
        millisecond = t.milliseconds
        seconds = millisecond // 1000
        ss = seconds % 60
        minutes = seconds // 60
        mm = minutes % 60
        hours = minutes // 60
        hh = hours % 24
        fff = millisecond % 1000
        if tc_mode is TCMode.TC:
            if timebase is None:
                timebase = Timebase(24, False)
            rate = timebase.fps
            fff = int((millisecond / 1000) * rate) % rate
        return TimeCodeInfo(
            hh=hh, mm=mm, ss=ss, ff=fff, mode=tc_mode, timebase=timebase
        )

    @property
    def milliseconds(self) -> int:
        fff = self.ff
        if self.mode is TCMode.Code:
            rate = 24 if self.timebase is None else self.timebase.frame_rate
            fff = self.ff / rate * 1000
        return int(
            round(
                self.hh * 60 * 60 * 1000 + self.mm * 60 * 1000 + self.ss * 1000 + fff)
        )

    @property
    def segments(self):
        return tuple(self.hh, self.mm, self.ss, self.ff)


class TimeCode:
    __STR_TEMPLATE = {
        TCMode.Code: "{0[0]:0>2}{sep}{0[1]:0>2}{sep}{0[2]:0>2}{sep1}{0[3]:0>2}",
        TCMode.Stamp: "{0[0]:0>2}{sep}{0[1]:0>2}{sep}{0[2]:0>2}{sep1}{0[3]:0>3}",
    }

    __DEF_SEP = {
        TCMode.Code: ":;",
        TCMode.Stamp: ".,",
    }

    def __init__(
            self,
            input_data: Union[Time | int | float | str | TimeCodeInfo],
            tc_mode: TCMode = TCMode.Code,
            timebase: Timebase = None,
            custom_seps: str = None,
    ) -> None:
        """初始化时间码对象

        Args:
            input_time (Union[Time  |  int  |  float  |  str]): 支持多种格式的自动识别
            timebase (Timebase, optional): 指定时基，如果为Stamp模式则不需要. Defaults to None.
            tc_mode (TCMode, optional): 时间码模式，是帧数模式还是毫秒模式. Defaults to TCMode.Code.
            custom_seps (str, optional): 应该提供字符串，其中的第一个字符用作普通的分隔符，
            第二个字符（如果有的话）用作丢帧的分隔符. Defaults to None.
        """
        if isinstance(input_data, TimeCodeInfo):
            self._time = Time(input_data.milliseconds)
            self._mode = input_data.mode
            self._timebase = input_data.timebase
        if isinstance(input_data, str) and TimeCodeInfo.is_tc(input_data):
            info = TimeCodeInfo.from_timecode(input_data)
            self._time = Time(info.milliseconds)
        elif isinstance(input_data, Time):
            self._time = input_data
        elif isinstance(input_data, int | float):
            self._time = Time(input_data)
        else:
            self._time = Time()

        self._timebase = timebase if timebase is not None else None
        self._mode = tc_mode
        self._custom_seps = str(
            custom_seps) if custom_seps is not None else None

    @property
    def seps(self) -> str:
        if self._custom_seps:
            return self._custom_seps
        return self.__DEF_SEP[self._mode]

    @property
    def time(self) -> Time:
        return self._time

    def render(self) -> str:
        info = TimeCodeInfo.from_time(self._time, self._mode, self._timebase)
        seps = self.seps
        sep = seps[0]
        sep1 = sep
        if self._timebase.drop_frame and len(seps) > 1:
            sep1 = seps[1]

        return self.__STR_TEMPLATE[self._mode].format(
            info.segements, sep=sep, sep1=sep1
        )

    def __str__(self):
        return self.render()

    def __repr__(self):
        return "TimeCode{0}{1}".format(
            self._time, "DF" if self._timebase.drop_frame else ""
        )



from typing import Union

from ..core import DataPackage, Time, TimeRangeSupport


class Item(TimeRangeSupport):
    def __init__(self, start: Union[int, Time] = Time(),
                 duration: Union[Time, int] = Time()) -> None:
        super().__init__()
        self._start = Time(start)
        self._duration = Time(duration)
        self._data = DataPackage()

    @classmethod
    def from_time_range(cls, time_range: TimeRangeSupport):
        assert isinstance(time_range, TimeRangeSupport)
        return cls(time_range.start, time_range.duration)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        value = Time(value)
        self._start = value

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, value):
        value = Time(value)
        self._duration = value

    @property
    def end(self):
        return self.start + self.duration

    @end.setter
    def end(self, value):
        value = Time(value)
        self.duration = value - self.start

    @property
    def data(self):
        return self._data



from abc import ABC, abstractmethod
from typing import Union

from ..cx_subtitle import StaticSubtitle


class SubtitleProcessor(ABC):

    @abstractmethod
    def __call__(self,
                 subtitle: StaticSubtitle) -> Union[StaticSubtitle | None]:
        return subtitle



from cx_studio.core import TimeCode
from .cx_subtitle_saver import SubtitleSaver
from ..cx_subtitle import StaticSubtitle


class SrtSaver(SubtitleSaver):
    SUBTITLE_PATTERN = '{number}\n{start} --> {end}\n{content}\n\n'

    def __init__(self, filename, encoding=None, newline='\n'):
        super().__init__(filename, encoding)
        self._newline = newline or ''
        self._file = None
        self._number = 1

    def open(self):
        self._file = open(self._filename, 'wt', encoding=self.encoding)
        return True

    def close(self):
        if self._file:
            self._file.flush()
            self._file.close()
        return True

    def _compile_content(self, number: int, subtitle: StaticSubtitle):
        start = TimeCode(subtitle.start)
        end = TimeCode(subtitle.end)
        contents = subtitle.content.splitlines()
        return self.SUBTITLE_PATTERN.format(
            number=number,
            start=start, end=end,
            content=self._newline.join(contents)
        )

    def write_subtitle(self, subtitle: StaticSubtitle):
        compiled_subtitle = self._compile_content(self._number, subtitle)
        self._file.write(compiled_subtitle)
        self._number += 1



from dataclasses import dataclass


@dataclass(frozen=True)
class Timebase:
    frame_rate: int = 24
    drop_frame: bool = False

    @property
    def milliseconds_per_frame(self) -> int:
        return round(1.0 / float(self.frame_rate) * 1000)

    def is_valid(self) -> bool:
        if self.frame_rate <= 0:
            return False
        return True



from typing import Callable

from .cx_subtitle_processor import SubtitleProcessor
from ..cx_subtitle import StaticSubtitle


class SubtitleChainProcessor(SubtitleProcessor):
    def __init__(self, *processors: list[Callable]):
        super().__init__()
        self.processors = []
        self.processors += processors

    def append(self, processor: Callable[[StaticSubtitle], StaticSubtitle]):
        self.processors.append(processor)
        return self

    def __call__(self, subtitle: StaticSubtitle):
        result = subtitle
        for p in self.processors:
            result = p(result)
            if result is None:
                break
        return result



from .cx_datapackage import DataPackage
from .cx_membercaches import MemberCaches
from .cx_time import Time
from .cx_timebase import Timebase
from .cx_timecode import TCMode, TimeCode
from .cx_timerangesupport import TimeRangeSupport



from numbers import Number
from typing import SupportsInt


class Time:
    SECOND_MS: int = 1000
    MINUTE_MS: int = SECOND_MS * 60
    HOUR_MS: int = MINUTE_MS * 60

    def __init__(self, milliseconds=0) -> None:
        if not isinstance(milliseconds, Number | Time):
            raise NotImplemented("Input is not a acceptable number.")

        self._ms: int = int(milliseconds)

    @property
    def milliseconds(self) -> int:
        return self._ms

    @property
    def seconds(self) -> float:
        return self._ms / self.SECOND_MS

    @property
    def hours(self) -> float:
        return self._ms / self.HOUR_MS

    @property
    def minutes(self) -> float:
        return self._ms / self.MINUTE_MS

    def __eq__(self, other):
        if isinstance(other, Time):
            return self._ms == other._ms
        elif isinstance(other, SupportsInt):
            return self._ms == int(other)
        else:
            raise NotImplemented("Comparing type not supported.")

    def __hash__(self) -> int:
        return hash(f"CXTIME:{self._ms}")

    def __lt__(self, other):
        if isinstance(other, Time):
            return self._ms < other._ms
        elif isinstance(other, SupportsInt):
            return self._ms < int(other)
        else:
            raise NotImplemented("Comparing type not supported.")

    def __le__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return self.__eq__(other) or self.__gt__(other)

    def __add__(self, other):
        if isinstance(other, Time):
            return Time(self._ms + other._ms)
        elif isinstance(other, SupportsInt):
            return Time(self._ms + int(other))
        else:
            raise NotImplemented()

    def __sub__(self, other):
        if isinstance(other, Time):
            return Time(self._ms - other._ms)
        elif isinstance(other, SupportsInt):
            return Time(self._ms - int(other))
        else:
            raise NotImplemented()

    def __mul__(self, other):
        if isinstance(other, SupportsInt):
            return Time(self._ms * int(other))
        else:
            raise NotImplemented()

    def __truediv__(self, other):
        if isinstance(other, SupportsInt):
            return Time(round(self._ms / int(other)))
        else:
            raise NotImplemented()

    def __repr__(self):
        return f"CXTime[{self._ms}]"

    def __int__(self):
        return self._ms



import dataclasses
import re

from .cx_subtitle_processor import SubtitleProcessor
from ..cx_subtitle import StaticSubtitle


class SubtitlePrettifier(SubtitleProcessor):
    def __init__(self, **kwargs):
        super(SubtitlePrettifier, self).__init__()
        self.replacements = {'\t': ' '}

        self.remove_xml_tags = True

        self.normal_strip = True
        self.quotes_strip = True
        self.extra_strip = ''

        self.remove_empty_lines = True
        self.shrink_long_spaces = True
        self.shrink_long_quotes = True

        self.remove_empty_subtitles = True
        self.remove_zero_duration_subtitles = True

        self.__dict__.update(kwargs)

    def __basic_content_cleaning(self, content: str):
        result = content
        for k, v in self.replacements.items():
            result = re.sub(k, v, result)
        result = re.sub(r'\r\n', r'\n', result)
        return result

    @staticmethod
    def _remove_xml_tags(content: str):
        return re.sub(
            r'<(\w+)>([^<>]*)</\1>',
            r'\2',
            content
        )

    def _strip(self, content: str):
        result = str(content)
        while True:
            current = result
            if self.normal_strip:
                current = current.strip()
            if self.quotes_strip:
                current = current.strip('\'"‘’“”')
            if self.extra_strip:
                current = current.strip(self.extra_strip)
            if current == result:
                break
            result = current
        return result

    @staticmethod
    def _remove_empty_lines(content: str):
        return re.sub(r'\n\s+', r'\n', content)

    @staticmethod
    def _shrink_long_spaces(content: str):
        return re.sub(r' +', r' ', content)

    @staticmethod
    def _shrink_long_quotes(content: str):
        return re.sub(r'([\'\"‘“”’])+', r'\1', content)

    def is_subtitle_legal(self, subtitle: StaticSubtitle) -> bool:
        if self.remove_empty_subtitles and not subtitle.content:
            return False
        if self.remove_zero_duration_subtitles and subtitle.duration() == 0:
            return False
        return True

    def __call__(self, subtitle: StaticSubtitle):
        content = self.__basic_content_cleaning(subtitle.content)
        content = self._strip(content)
        if self.remove_xml_tags:
            content = self._remove_xml_tags(content)
        if self.remove_empty_lines:
            content = self._remove_empty_lines(content)
        if self.shrink_long_spaces:
            content = self._shrink_long_spaces(content)
        if self.shrink_long_quotes:
            content = self._shrink_long_quotes(content)
        result = dataclasses.replace(subtitle, content=content)
        return result if self.is_subtitle_legal(result) else None



import os
from pathlib import Path
from typing import Iterable


def normalize_path(path: Path, anchor=Path(".")) -> Path:
    path = Path(path)
    anchor = Path(anchor)
    if path.is_absolute():
        return path.absolute()

    t = str(path)
    if t.startswith("~"):
        return Path(t.replace("~", str(Path.home()))).absolute()

    return (anchor / path).resolve(strict=False)


def normalize_suffix(suffix: str, with_dot=True) -> str:
    """检查扩展名是否带点"""
    s = str(suffix).strip().strip(".").lower()
    return "." + s if with_dot else s


def force_suffix(source: Path, suffix: str, ignore_dir=True):
    if not source:
        return None
    source = Path(source)
    if source.is_dir() and ignore_dir:
        return source
    suffix = normalize_suffix(suffix)
    return source if source.suffix == suffix else source.with_suffix(suffix)


def ensure_folder(source: Path):
    source = normalize_path(source)
    if source.is_file():
        source = source.parent
    return source


def quote_path(path, quote_char='"') -> str:
    path = str(path)
    return f"{quote_char}{path}{quote_char}" if " " in path else path


def is_file_in_dir(file, folder) -> bool:
    f = str(Path(file).resolve().absolute())
    d = str(Path(folder).resolve().absolute())
    return f in d


def is_executable(cmd: Path) -> bool:
    cmd = Path(cmd)
    return cmd.exists() and os.access(cmd, os.X_OK)


def ensure_paths(sources: Iterable) -> list[Path]:
    if not sources:
        sources = []
    for x in sources:
        if isinstance(x, Iterable):
            yield from ensure_paths(x)
        elif isinstance(x, Path):
            yield x
        else:
            yield Path(str(x))




import itertools
import os
from pathlib import Path
from cx_studio.core import MemberCaches
from .path_utils import ensure_paths, is_executable


class CommandChecker:
    def __init__(
        self,
        executable: str,
        include_pwd=True,
        extra_paths: list[Path] = None,
        extra_suffixes: list[str] = None,
    ):
        self._source = str(executable)
        self._include_pwd = include_pwd
        self._extra_paths = [x for x in ensure_paths(extra_paths)]
        self._extra_suffixes = extra_suffixes if extra_suffixes else []

    def _possible_sources(self):
        possible_suffixes = [".exe", ".com"]
        yield self._source
        for suffix in possible_suffixes:
            yield self._source + suffix
        for x in self._extra_suffixes:
            suffix = str(x)
            if not suffix.startswith("."):
                suffix = "." + suffix
            yield self._source + suffix

    @staticmethod
    def os_paths():
        for p in os.environ["PATH"].split(os.pathsep):
            yield Path(p)

    def additional_paths(self):
        for p in self._extra_paths:
            if p.is_dir():
                yield p
        if self._include_pwd:
            yield Path.cwd()

    def search_paths(self):
        yield from self.os_paths()
        yield from self.additional_paths()

    def executable(self) -> str:
        if os.path.isabs(self._source) and is_executable(Path(self._source)):
            return self._source

        for folder, file in itertools.product(
            self.os_paths(), self._possible_sources()
        ):
            cmd = Path(folder) / file
            if is_executable(cmd):
                return self._source

        for folder, file in itertools.product(
            self.additional_paths(), self._possible_sources()
        ):
            cmd = Path(folder) / file
            if is_executable(cmd):
                return cmd.resolve()

        return ""

    def absolute(self) -> str:
        result = self.executable()
        if result:
            return str(Path(result).resolve())
        return ""



import dataclasses
from numbers import Number
from typing import Union

from cx_studio.core import Time, Timebase
from .cx_subtitle_processor import SubtitleProcessor
from ..cx_subtitle import StaticSubtitle


class SubtitleFrameRateShifter(SubtitleProcessor):
    def __init__(self, from_rate: Union[Number, Timebase],
                 to_rate: Union[Number,
                 Timebase]):
        super().__init__()
        self._from_rate = self.__unpack_rate(from_rate)
        self._to_rate = self.__unpack_rate(to_rate)

    @staticmethod
    def __unpack_rate(rate: Union[Number, Timebase]) -> Timebase:
        if isinstance(rate, Number):
            return Timebase(int(rate))
        if isinstance(rate, Timebase):
            return rate
        raise NotImplementedError('Unsupported frame rate format')

    def transform_time_with_fixed_frames(self, time: Time) -> Time:
        frames = int(round(time.milliseconds /
                           self._from_rate.milliseconds_per_frame))
        new_milliseconds = frames * self._to_rate.milliseconds_per_frame
        return Time(new_milliseconds)

    def __call__(self, subtitle: StaticSubtitle):
        start = self.transform_time_with_fixed_frames(subtitle.start)
        end = self.transform_time_with_fixed_frames(subtitle.end)
        return dataclasses.replace(subtitle, start=start, end=end)



from .cx_track import Track
from ..core import DataPackage, Time, TimeRangeSupport


class Timeline(TimeRangeSupport):
    def __init__(self) -> None:
        super().__init__()
        self._tracks: list[Track] = [Track()]
        self._data = DataPackage()

    @property
    def data(self):
        return self._data

    @property
    def tracks(self):
        return self._tracks

    @property
    def start(self):
        return Time(0)

    @property
    def duration(self):
        return max(t.duration for t in self._tracks)



from abc import ABC, abstractmethod

from .cx_time import Time


class TimeRangeSupport(ABC):
    @property
    @abstractmethod
    def start(self) -> Time:
        pass

    @property
    @abstractmethod
    def duration(self) -> Time:
        pass

    @property
    def end(self) -> Time:
        return self.start + self.duration

    def intersects(self, other) -> bool:
        assert isinstance(other, TimeRangeSupport)
        return self.start <= other.end and self.end >= other.start

    def contains_time(self, time: Time) -> bool:
        assert isinstance(time, Time)
        return self.start <= time <= self.end



from .cx_subtitle_chain_processor import SubtitleChainProcessor
from .cx_subtitle_frame_rate_shifter import SubtitleFrameRateShifter
from .cx_subtitle_prettifier import SubtitlePrettifier
from .cx_subtitle_processor import SubtitleProcessor



import re
from enum import StrEnum

from cx_studio.core import TCMode, Time, TimeCode
from .cx_subtitle_loader import SubtitleLoader
from ..cx_subtitle import StaticSubtitle


class SrtLoader(SubtitleLoader):
    class State(StrEnum):
        ExpectingNumber = "number"
        ExpectingTimecode = "timecode"
        ExpectingContent = "content"

    PATTERN_NUMBER = re.compile(r'^\d+$')
    PATTERN_TIME = re.compile(
        r'^(\d\d:\d\d:\d\d[,.]\d\d\d) --> (\d\d:\d\d:\d\d[,.]\d\d\d)$')

    def __init__(self, filename, encoding=None, newline='\n'):
        super().__init__(filename, encoding)
        self._newline = newline or ''
        self._file = None
        self._state = self.State.ExpectingNumber

    def open(self):
        self._file = open(self._filename, 'rt', self.encoding)
        return True

    def close(self):
        if self._file:
            self._file.close()

    def subtitles(self):
        start, end, content = Time(), Time(), []

        for a in self._file:
            line = str(a).strip()
            match self._state:
                case self.State.ExpectingNumber:
                    if self.PATTERN_NUMBER.fullmatch(line):
                        self._state = self.State.ExpectingTimecode
                case self.State.ExpectingTimecode:
                    match = self.PATTERN_TIME.fullmatch(line)
                    if match:
                        start = TimeCode(match.group(1), TCMode.Stamp).time
                        end = TimeCode(match.group(2), TCMode.Stamp).time
                        self._state = self.State.ExpectingContent
                case self.State.ExpectingContent:
                    if line:
                        content.append(line)
                    else:
                        yield StaticSubtitle(start=start,
                                             end=end,
                                             content=self._newline.join(
                                                 content))
                        self._state = self.State.ExpectingNumber



import functools


class MemberCaches:
    def __init__(self):
        self._cache = {}

    def cache(self, key: str):
        def actual_decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if key not in self._cache:
                    self._cache[key] = func(*args, **kwargs)
                return self._cache[key]

            return wrapper

        return actual_decorator

    def clear(self, key: str = None):
        if key is None:
            self._cache.clear()
        elif key in self._cache:
            del self._cache[key]

    def __getitem__(self, item):
        return self._cache[item]

    def __setitem__(self, key, value):
        self._cache[key] = value

    def __delitem__(self, key):
        del self._cache[key]

    def __len__(self):
        return len(self._cache)

    def __iter__(self):
        return self._cache.__iter__()



from ..core import DataPackage, Time, TimeRangeSupport


class Track(TimeRangeSupport):
    def __init__(self) -> None:
        super().__init__()
        self._items: list[TimeRangeSupport] = []
        self._data = DataPackage()

    @property
    def start(self) -> Time:
        return Time(0)

    @property
    def duration(self) -> Time:
        if len(self.items) == 0:
            return Time(0)
        else:
            return self._items[-1].end

    @property
    def items(self) -> list[TimeRangeSupport]:
        return self._items

    @property
    def data(self):
        return self._data

    def __getitem__(self, item):
        assert isinstance(item, int)
        return self._items[item]

    def __setitem__(self, key, value):
        assert isinstance(key, int)
        assert isinstance(value, TimeRangeSupport)
        self._items[key] = value

    def __delitem__(self, key):
        assert isinstance(key, int)
        del self._items[key]

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return self._items.__iter__()

    def find_insert_point(self, item):
        if len(self._items) == 0:
            return 0

        start_time = item.start if isinstance(item, TimeRangeSupport) else Time(
            item)
        for i in range(0, len(self._items)):
            if self._items[i].start > start_time:
                return i
        return len(self._items)

    def available_for(self, item: TimeRangeSupport):
        assert isinstance(item, TimeRangeSupport)
        for x in self._items:
            if x.intersects(item):
                return False
        return True

    def sort(self):
        self._items.sort(key=lambda x: x.start)



import re

from cx_studio.core import Time
from .cx_subtitle_loader import SubtitleLoader
from ..cx_subtitle import StaticSubtitle


class TxtLoader(SubtitleLoader):
    DURATION_PER_CHAR = 120
    MIN_DURATION_PER_LINE = 2000
    GAP_BETWEEN_ITEMS = 0
    TS_PATTERN = r'\s*\d{2}:\d{2}:\d{2}.\d{3}\s*'

    def __init__(self, filename, encoding=None,
                 duration_per_char: int = None,
                 min_duration_per_line: int = None,
                 gap_between_items: int = None):
        super().__init__(filename, encoding)
        self._file = None
        self._encoding = encoding
        self.duration_per_char = duration_per_char or self.DURATION_PER_CHAR
        self.gap_between_items = gap_between_items or self.GAP_BETWEEN_ITEMS
        self.min_duration_per_line = (
                min_duration_per_line or self.MIN_DURATION_PER_LINE)
        self._prev_time = Time()

    def open(self):
        self._file = open(self._filename, 'r', encoding=self.encoding)
        return True

    def close(self):
        if self._file:
            self._file.close()
        return True

    def subtitles(self):
        for line in self._file:
            content = re.sub(self.TS_PATTERN, '', line)
            duration = max(len(content) * self.duration_per_char,
                           self.min_duration_per_line)
            yield StaticSubtitle(
                start=self._prev_time,
                end=self._prev_time + Time(duration),
                content=content
            )
            self._prev_time += duration + self.gap_between_items



from cx_studio.core import TimeCode
from .cx_subtitle_saver import SubtitleSaver
from ..cx_subtitle import StaticSubtitle


class TxtSaver(SubtitleSaver):
    def __init__(self, filename, encoding=None, newline=None,
                 time_pattern=None):
        super().__init__(filename, encoding)
        self._newline = newline or ''
        self._time_pattern = time_pattern
        self._file = None

    def _compile_content(self, subtitle: StaticSubtitle):
        _time = ''
        if self._time_pattern:
            start_tc = TimeCode(subtitle.start)
            end_tc = TimeCode(subtitle.end)
            duration_tc = TimeCode(subtitle.duration())
            _time = str(self._time_pattern).format(start=start_tc, end=end_tc,
                                                   duration=duration_tc)
        _contents = subtitle.content.splitlines()
        return f'{_time}{self._newline.join(_contents)}\n'

    def open(self):
        self._file = open(self._filename, 'wt', encoding=self.encoding)
        return True

    def close(self):
        if self._file:
            self._file.flush()
            self._file.close()
        return True

    def write_subtitle(self, subtitle: StaticSubtitle):
        line = self._compile_content(subtitle)
        self._file.write(line)



class DataPackage:
    """
    DataPackage 是一个用于存储和访问嵌套数据的类。

    它允许使用点号（`.`）作为键来访问和设置数据，类似于 JavaScript 对象。
    数据存储在一个嵌套的字典结构中，可以存储基本数据类型、列表和自身。
    """

    def __init__(self, **kwargs):
        self._data = {}
        for k, v in kwargs.items():
            self._data[str(k)] = DataPackage.__check_value(v)

    def __repr__(self):
        return repr(self._data)

    @staticmethod
    def __check_value(value):
        if isinstance(value, dict):
            return DataPackage(**value)
        if isinstance(value, list):
            return [DataPackage.__check_value(x) for x in value]
        return value

    def _get_value(self, child_key, *keys):
        child = self._data.get(child_key)
        if not keys:
            return child
        if isinstance(child, DataPackage):
            return child._get_value(*keys)
        return None

    def _set_value(self, child_key, *keys, value=None):
        if not keys:
            self._data[child_key] = value
            return

        if child_key not in self._data:
            self._data[child_key] = DataPackage()

        child = self._data[child_key]
        if isinstance(child, DataPackage):
            child._set_value(*keys, value=value)
            return

        raise KeyError("Invalid key path:", child_key, keys)

    def _del_value(self, child_key, *keys):
        if not keys:
            if child_key not in self._data:
                raise KeyError("Invalid key path", child_key, keys)
            del self._data[child_key]
        elif isinstance(self._data[child_key], DataPackage):
            self._data[child_key]._del_value(*keys)
        else:
            raise KeyError('Invalid key path', child_key, keys)

    def __getitem__(self, item):
        keys = str(item).split(".")
        return self._get_value(*keys)

    def __setitem__(self, key, value):
        keys = str(key).split(".")
        return self._set_value(*keys, value=DataPackage.__check_value(value))

    def __delitem__(self, key):
        keys = str(key).split('.')
        self._del_value(*keys)

    def __getattr__(self, item):
        keys = str(item).split(".")
        return self._get_value(*keys)

    def __setattr__(self, key, value):
        s_key = str(key)
        if s_key.startswith("_"):
            self.__dict__[key] = value
        # elif '.' in s_key:
        #     self._set_value(*(s_key.split('.')), value=DataPackage.__check_value(value))
        else:
            self._data[s_key] = DataPackage.__check_value(value)

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def get(self, path: str, default_value=None):
        result = default_value
        try:
            result = self._get_value(*path.split("."))
        except NotImplemented:
            pass
        return result
